"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const assync_1 = require("assync");
const deps_1 = require("./deps");
const plugins_1 = require("./plugins");
const builtin_1 = require("./plugins/builtin");
const topic_1 = require("./topic");
class CommandManager {
    constructor(config) {
        this.config = config;
        this.managers = [new builtin_1.Builtin(this.config), new plugins_1.Plugins(this.config)];
        this.debug = require('debug')('cli:command');
        this.hooks = new deps_1.default.Hooks(this.config);
    }
    async run(argv) {
        await this.hooks.run('init');
        const id = argv[2];
        await this.load();
        if (this.cmdAskingForHelp(argv))
            return this.help(argv.slice(3));
        let cmd = this.result.findCommand(id);
        if (!cmd) {
            let topic = await this.findTopic(id);
            if (topic)
                return this.help([id]);
            else
                return this.notFound(id);
        }
        await this.hooks.run('prerun', { argv, Command: cmd });
        this.debug('running %s', cmd.id);
        const result = await cmd.run(argv);
        this.debug('exited normally');
        return result;
    }
    async help(argv) {
        deps_1.default.Help.run(argv, this.config);
    }
    async notFound(id) {
        deps_1.default.NotFound.run([id], this.config);
    }
    async commands() {
        await this.load();
        return this.result.allCommands;
    }
    async rootCommands() {
        await this.load();
        return Object.values(this.result.commands);
    }
    async findTopic(id) {
        await this.load();
        return this.result.findTopic(id);
    }
    async findCommand(id, must) {
        await this.load();
        let cmd = this.result.findCommand(id);
        if (!cmd && must)
            throw new Error(`${id} not found`);
        return cmd;
    }
    async topics() {
        await this.load();
        return this.result.allTopics;
    }
    async load() {
        if (this.result)
            return;
        this.debug('load');
        this.result = new topic_1.RootTopic();
        await this.hooks.run('init');
        let managers = await this.submanagers();
        this.debug('loading all managers');
        let loads = managers.filter(m => m.load).map(m => m.load());
        for (let r of loads)
            this.addResult(await r);
    }
    async submanagers() {
        if (this._submanagers)
            return this._submanagers;
        const fetch = async (managers) => {
            const submanagers = await assync_1.default(managers)
                .filter(m => !!m.submanagers)
                .map(m => Promise.resolve(m.submanagers()).then(fetch))
                .flatMap();
            return [...(managers || []), ...(submanagers || [])];
        };
        this.debug('fetching command managers');
        this._submanagers = await fetch(this.managers);
        this.debug('received command %d managers', this._submanagers.length);
        return this._submanagers;
    }
    addResult(result) {
        this.result.addTopics(result.topics);
        this.result.addCommands(result.commands);
    }
    cmdAskingForHelp(argv) {
        if (argv[2] === 'help')
            return true;
        for (let arg of argv) {
            if (['--help', '-h'].includes(arg))
                return true;
            if (arg === '--')
                return false;
        }
        return false;
    }
}
exports.CommandManager = CommandManager;
