"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const path = require("path");
const ts_lodash_1 = require("ts-lodash");
const deps_1 = require("../deps");
const core_1 = require("./core");
const link_1 = require("./link");
const user_1 = require("./user");
class Plugins {
    constructor(config) {
        this.config = config;
        this.debug = require('debug')('cli:plugins');
        this.core = new core_1.CorePlugins(this.config);
        this.user = new user_1.UserPlugins(this.config);
        this.link = new link_1.LinkPlugins(this.config);
    }
    async submanagers() {
        return [this.link, this.user, this.core];
    }
    async install(options) {
        await this.init();
        let name = options.type === 'user' ? options.name : await this.getLinkedPackageName(options.root);
        if (!options.force && (await this.pluginType(name))) {
            throw new Error(`${name} is already installed, run with --force to install anyway`);
        }
        if (options.type === 'link') {
            await this.link.install(options.root);
        }
        else {
            await this.user.install(name, options.tag);
        }
    }
    async update() {
        await this.user.update();
    }
    async uninstall(name) {
        const type = await this.pluginType(name);
        if (!type) {
            const linked = await this.link.findByRoot(name);
            if (linked) {
                name = linked.name;
            }
            else
                throw new Error(`${name} is not installed`);
        }
        if (type === 'user')
            await this.user.uninstall(name);
    }
    async list() {
        await this.init();
        return this.plugins;
    }
    async init() {
        if (this.plugins)
            return;
        const managers = ts_lodash_1.default.compact([this.link, this.user, this.core]);
        await Promise.all(managers.map(m => m.init()));
        const plugins = managers.reduce((o, i) => o.concat(i.plugins), []);
        this.plugins = ts_lodash_1.default.compact([...plugins, this.builtin]);
        await await this.plugins.map(p => p.load());
    }
    async getLinkedPackageName(root) {
        const pjson = await deps_1.default.file.fetchJSONFile(path.join(root, 'package.json'));
        return pjson.name;
    }
    pluginType(name) {
        const plugin = this.plugins.find(p => p.name === name);
        return plugin && plugin.type;
    }
}
exports.Plugins = Plugins;
