"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const fs = require("fs-extra");
const path = require("path");
const deps_1 = require("../deps");
const plugin_1 = require("./plugin");
function touch(f) {
    fs.utimesSync(f, new Date(), new Date());
}
function linkPJSON(root) {
    return deps_1.default.file.fetchJSONFile(path.join(root, 'package.json'));
}
async function getNewestJSFile(root) {
    let files = await deps_1.default.file.walk(root, {
        depthLimit: 10,
        filter: f => !['.git', 'node_modules'].includes(path.basename(f)),
    });
    return files.reduce((prev, f) => {
        if (f.stats.isDirectory())
            return prev;
        if (f.path.endsWith('.js') || f.path.endsWith('.ts')) {
            if (f.stats.mtime > prev)
                return f.stats.mtime;
        }
        return prev;
    }, new Date(0));
}
class LinkPlugins {
    constructor(config) {
        this.config = config;
        this.debug = require('debug')('cli:plugins:user');
    }
    async install(root) {
        await this.init();
        let downgrade = await this.lock.upgrade();
        this.debug('installing', root);
        const plugin = await this.loadPlugin(root);
        await plugin.init();
        await plugin.refresh(true);
        await plugin.load();
        await this.addPlugin(plugin.name, plugin.root);
        await downgrade();
    }
    async findByRoot(root) {
        await this.init();
        root = path.resolve(root);
        return this.plugins.find(p => path.resolve(p.root) === root);
    }
    async submanagers() {
        await this.init();
        return this.plugins;
    }
    async init() {
        if (this.plugins)
            return;
        this.debug('init');
        this.manifest = new deps_1.default.PluginManifest({
            name: 'link',
            file: path.join(this.config.dataDir, 'plugins', 'link.json'),
        });
        this.lock = new deps_1.default.Lock(this.config, this.manifest.file + '.lock');
        await this.migrate();
        const manifest = await this.manifest.get('plugins');
        this.plugins = await Promise.all(Object.values(manifest || {}).map(v => this.loadPlugin(v.root)));
        if (this.plugins.length)
            this.debug('plugins:', this.plugins.map(p => p.name).join(', '));
    }
    async migrate() {
        const linkedPath = path.join(this.config.dataDir, 'linked_plugins.json');
        if (!await deps_1.default.file.exists(linkedPath))
            return;
        let downgrade = await this.lock.upgrade();
        this.debug('migrating link plugins');
        let linked = await deps_1.default.file.readJSON(linkedPath);
        for (let root of linked.plugins) {
            const name = await deps_1.default.file.readJSON(path.join(root, 'package.json'));
            await this.addPlugin(name, root);
        }
        await deps_1.default.file.remove(linkedPath);
        await downgrade();
    }
    async addPlugin(name, root) {
        let plugins = (await this.manifest.get('plugins')) || {};
        plugins[name] = { root };
        await this.manifest.set('plugins', plugins);
        await this.manifest.save();
    }
    async loadPlugin(root) {
        return new LinkPlugin({
            config: this.config,
            root,
            pjson: await linkPJSON(root),
        });
    }
}
exports.LinkPlugins = LinkPlugins;
class LinkPlugin extends plugin_1.Plugin {
    constructor() {
        super(...arguments);
        this.type = 'link';
    }
    async resetCache() {
        await super.resetCache();
        await this.manifest.set('lastUpdated', new Date().toISOString());
        await this.manifest.save();
    }
    async refresh(force = false) {
        if (force)
            return this.updateNodeModules();
        let type = await this.refreshType();
        switch (type) {
            case 'node_modules':
                await this.updateNodeModules();
                break;
            case 'prepare':
                await this.prepare();
                break;
        }
    }
    async init() {
        await super.init();
        this.manifest = new deps_1.default.PluginManifest({
            name: 'link',
            file: path.join(this.config.dataDir, 'plugins', 'link', `${this.name}.json`),
        });
        this.lock = new deps_1.default.Lock(this.config, this.manifest.file + '.lock');
    }
    async refreshType() {
        if (await this.updateNodeModulesNeeded())
            return 'node_modules';
        if (await this.prepareNeeded())
            return 'prepare';
    }
    async updateNodeModulesNeeded() {
        let modules = path.join(this.root, 'node_modules');
        if (!await deps_1.default.file.exists(modules))
            return true;
        let modulesInfo = await fs.stat(modules);
        let pjsonInfo = await fs.stat(path.join(this.root, 'package.json'));
        return modulesInfo.mtime < pjsonInfo.mtime;
    }
    async prepareNeeded() {
        const main = this.pjson.main;
        if (main && !await deps_1.default.file.exists(path.join(this.root, main)))
            return true;
        return (await this.lastUpdated()) < (await getNewestJSFile(this.root));
    }
    async updateNodeModules() {
        this.debug('update node modules');
        const yarn = new deps_1.default.Yarn({ config: this.config, cwd: this.root });
        await yarn.exec();
        touch(path.join(this.root, 'node_modules'));
        await this.resetCache();
    }
    async prepare() {
        this.debug('prepare');
        const { scripts } = this.pjson;
        if (!scripts || !scripts.prepare)
            return;
        const yarn = new deps_1.default.Yarn({ config: this.config, cwd: this.root });
        await yarn.exec(['run', 'prepare']);
        await this.resetCache();
    }
    async lastUpdated() {
        const lastUpdated = await this.manifest.get('lastUpdated');
        return lastUpdated ? new Date(lastUpdated) : new Date(0);
    }
}
exports.LinkPlugin = LinkPlugin;
