"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const fs = require("fs-extra");
const deps_1 = require("../deps");
class PluginManifest {
    constructor(opts) {
        this.needsSave = false;
        this.file = opts.file;
        this.invalidate = opts.invalidate;
        this.name = opts.name;
        this.debug = require('debug')(`cli:manifest:${this.name}`);
    }
    async save() {
        await this.init();
        if (!this.needsSave)
            return;
        this.needsSave = false;
        this.debug('saving');
        if (!await this.canWrite()) {
            throw new Error('manifest file modified, cannot save');
        }
        await deps_1.default.file.outputJSON(this.file, this.body);
    }
    async fetch(key, fn) {
        await this.init();
        let v = await this.get(key);
        if (!v) {
            this.debug('fetching', key);
            await this.set(key, await fn());
        }
        return await this.get(key);
    }
    async get(key) {
        await this.init();
        return this.body.manifest[key];
    }
    async set(key, v) {
        if (!key)
            throw new Error('key is empty');
        await this.init();
        this.body.manifest[key] = v;
        this.needsSave = true;
        return this.body.manifest[key];
    }
    async reset() {
        await deps_1.default.file.remove(this.file);
        delete this.body;
        this.needsSave = false;
    }
    async init() {
        await this.saving;
        if (this.body)
            return this.body;
        this.body = (await this.read()) || {
            invalidate: this.invalidate,
            manifest: {},
        };
    }
    async read() {
        try {
            this.mtime = await this.getLastUpdated();
            let body = await fs.readJSON(this.file);
            if (body.invalidate !== this.invalidate) {
                this.debug('manifest version mismatch');
                return;
            }
            if (!body.manifest)
                this.body.manifest = {};
            return body;
        }
        catch (err) {
            if (err.code === 'ENOENT') {
                this.debug(err);
            }
            else
                throw err;
        }
    }
    async canWrite() {
        if (!this.mtime)
            return true;
        return (await this.getLastUpdated()) === this.mtime;
    }
    async getLastUpdated() {
        try {
            const stat = await deps_1.default.file.stat(this.file);
            return stat.mtime.getTime();
        }
        catch (err) {
            if (err.code !== 'ENOENT')
                throw err;
        }
    }
}
exports.PluginManifest = PluginManifest;
