"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const path = require("path");
const deps_1 = require("../deps");
class Plugin {
    constructor(opts) {
        this.config = opts.config;
        this.root = opts.root;
    }
    async load() {
        await this.init();
        const results = {
            commands: await this.commands(),
        };
        if (this.cache.needsSave) {
            let downgrade = await this.lock.write();
            await this.cache.save();
            await downgrade();
        }
        return results;
    }
    async yarnNodeVersion() {
        try {
            let f = await deps_1.default.file.readJSON(path.join(this.root, '..', '.yarn-integrity'));
            return f.nodeVersion;
        }
        catch (err) {
            if (err.code !== 'ENOENT')
                throw err;
        }
    }
    async resetCache() {
        await this.init();
        await this.cache.reset();
    }
    async init() {
        this.pjson = this.pjson || (await deps_1.default.file.fetchJSONFile(path.join(this.root, 'package.json')));
        if (!this.pjson['cli-engine'])
            this.pjson['cli-engine'] = {};
        this.name = this.name || this.pjson.name;
        this.version = this.version || this.pjson.version;
        let cacheKey = [this.config.version, this.version].join(path.sep);
        let cacheFile = path.join(this.config.cacheDir, 'plugins', [this.type, this.name + '.json'].join(path.sep));
        this.cache = new deps_1.default.PluginManifest({ file: cacheFile, invalidate: cacheKey, name: this.name });
        this.lock = new deps_1.default.Lock(this.config, cacheFile + '.lock');
        this.debug = require('debug')(`cli:plugins:${[this.type, this.name, this.version].join(':')}`);
        this.debug('init');
    }
    async findCommand(id, must = false) {
        let cmd = await this.findCommandInModule(id);
        if (!cmd)
            cmd = await this.findCommandInDir(id);
        if (cmd) {
            cmd = this.addPluginToCommand(cmd);
            return cmd;
        }
        if (must)
            throw new Error(`${id} not found`);
    }
    async commands() {
        const cache = await this.cache.fetch('commands', async () => {
            this.debug('fetching commands');
            const commands = await deps_1.default
                .assync([this.commandsFromModule(), this.commandsFromDir()])
                .flatMap();
            if (!commands.length)
                throw new Error(`${this.name} has no commands`);
            return Promise.all(commands);
        });
        return cache.map(c => (Object.assign({}, c, { run: async (argv) => {
                let cmd = await this.findCommand(c.id, true);
                if (!c._version || c._version === '0.0.0') {
                    // this.debug('legacy cli-engine-command version', c._version)
                    return cmd.run(Object.assign({}, this.config, { argv: argv.slice(4) }));
                }
                else if (deps_1.default.semver.lt(c._version || '', '11.0.0')) {
                    // this.debug(`legacy cli-engine-command version`, c._version)
                    return cmd.run(Object.assign({}, this.config, { argv: argv.slice(2) }));
                }
                else {
                    return cmd.run(argv.slice(3), this.config);
                }
            } })));
    }
    get commandsDir() {
        let d = this.pjson['cli-engine'].commandsDir;
        if (d)
            return path.join(this.root, d);
    }
    commandPath(id) {
        if (!this.commandsDir)
            throw new Error('commandsDir not set');
        return require.resolve(path.join(this.commandsDir, id.split(':').join(path.sep)));
    }
    async commandsFromModule() {
        const m = await this.fetchModule();
        if (!m)
            return [];
        return deps_1.default.assync(m.commands).map(c => this.commandInfoFromICommand(c));
    }
    async commandsFromDir() {
        const ids = await this.commandIDsFromDir();
        return deps_1.default
            .assync(ids)
            .map(id => this.findCommandInDir(id))
            .map(c => this.commandInfoFromICommand(c));
    }
    async commandInfoFromICommand(icommand) {
        return {
            _version: icommand._version,
            description: icommand.description,
            usage: icommand.usage,
            id: icommand.id,
            hidden: icommand.hidden,
            aliases: icommand.aliases || [],
            help: await icommand.buildHelp(this.config),
            helpLine: await icommand.buildHelpLine(this.config),
        };
    }
    findCommandInDir(id) {
        return deps_1.default.util.undefault(require(this.commandPath(id)));
    }
    async findCommandInModule(id) {
        const m = await this.fetchModule();
        if (!m)
            return;
        return m.commands.find(c => c.id === id);
    }
    addPluginToCommand(cmd) {
        cmd.plugin = {
            type: this.type,
            root: this.root,
            name: this.name,
            version: this.version,
        };
        return cmd;
    }
    commandIDsFromDir() {
        const d = this.commandsDir;
        if (!d)
            return Promise.resolve([]);
        return new Promise((resolve, reject) => {
            let ids = [];
            deps_1.default
                .klaw(d, { depthLimit: 10 })
                .on('data', f => {
                if (!f.stats.isDirectory() &&
                    (f.path.endsWith('.js') || (f.path.endsWith('.ts') && this.type === 'builtin')) &&
                    !f.path.endsWith('.d.ts') &&
                    !f.path.endsWith('.test.js')) {
                    let parsed = path.parse(f.path);
                    let p = path.relative(d, path.join(parsed.dir, parsed.name));
                    ids.push(p.split(path.sep).join(':'));
                }
            })
                .on('error', reject)
                .on('end', () => resolve(ids));
        });
    }
    async fetchModule() {
        if (!this.pjson.main)
            return;
        if (this._module)
            return this._module;
        return (this._module = (async () => {
            this.debug(`requiring ${this.name}@${this.version}`);
            const m = Object.assign({ commands: [], topics: [] }, require(path.join(this.root, this.pjson.main)));
            if (m.topic)
                m.topics.push(m.topic);
            m.commands = m.commands.map(deps_1.default.util.undefault);
            const hooks = new deps_1.default.Hooks(this.config);
            await hooks.run('plugins:parse', { module: m, pjson: this.pjson });
            let legacy = new deps_1.default.PluginLegacy(this.config);
            return legacy.convert(m);
        })());
    }
}
exports.Plugin = Plugin;
