"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const cli_ux_1 = require("cli-ux");
const path = require("path");
const deps_1 = require("../deps");
const plugin_1 = require("./plugin");
const yarn_1 = require("./yarn");
class UserPlugins {
    constructor(config) {
        this.config = config;
        this.debug = require('debug')('cli:plugins:user');
    }
    async submanagers() {
        await this.init();
        return this.plugins;
    }
    async update() {
        await this.init();
        if (this.plugins.length === 0)
            return;
        cli_ux_1.default.action.start(`${this.config.name}: Updating plugins`);
        const yarn = this.yarn();
        const manifest = await this.manifest.get('plugins');
        const packages = Object.entries(manifest || {}).map(([k, v]) => `${k}@${v}`);
        await yarn.exec(['upgrade', ...packages]);
    }
    async install(name, tag) {
        await this.init();
        let downgrade = await this.lock.upgrade();
        await this.createPJSON();
        const yarn = this.yarn();
        await yarn.exec(['add', `${name}@${tag}`]);
        const plugin = await this.loadPlugin(name, tag);
        await plugin.init();
        await plugin.resetCache();
        await plugin.load();
        await this.addPlugin(name, tag);
        await downgrade();
    }
    async uninstall(name) {
        const yarn = this.yarn();
        await yarn.exec(['remove', name]);
        await this.manifest.set(name, undefined);
    }
    async needsRefresh() {
        await this.init();
        if (!this.plugins.length)
            return false;
        let plugin = this.plugins[0];
        if ((await plugin.yarnNodeVersion()) !== process.version)
            return true;
        return false;
    }
    async refresh() {
        if (!await this.needsRefresh())
            return;
        const yarn = this.yarn();
        await yarn.exec();
        for (let p of this.plugins.map(p => p.resetCache()))
            await p;
    }
    async init() {
        if (this.plugins)
            return;
        this.debug('init');
        this.manifest = new deps_1.default.PluginManifest({
            name: 'user',
            file: path.join(this.config.dataDir, 'plugins', 'user.json'),
        });
        this.lock = new deps_1.default.Lock(this.config, this.manifest.file + '.lock');
        await this.migrate();
        const manifest = await this.manifest.get('plugins');
        this.plugins = await Promise.all(Object.entries(manifest || {}).map(([k, v]) => this.loadPlugin(k, v.tag)));
        if (this.plugins.length)
            this.debug('plugins:', this.plugins.map(p => p.name).join(', '));
        await this.refresh();
    }
    async loadPlugin(name, tag) {
        const pjson = await deps_1.default.file.fetchJSONFile(path.join(this.userPluginPath(name), 'package.json'));
        let p = new UserPlugin({
            pjson,
            tag,
            root: this.userPluginPath(name),
            config: this.config,
        });
        await p.init();
        return p;
    }
    userPluginPath(name) {
        return path.join(this.userPluginsDir, 'node_modules', name);
    }
    get userPluginsDir() {
        return path.join(this.config.dataDir, 'plugins');
    }
    get pjsonPath() {
        return path.join(this.userPluginsDir, 'package.json');
    }
    yarn() {
        return new yarn_1.default({ config: this.config, cwd: this.userPluginsDir });
    }
    async createPJSON() {
        if (!await deps_1.default.file.exists(this.pjsonPath)) {
            await deps_1.default.file.outputJSON(this.pjsonPath, { private: true, 'cli-engine': { schema: 1 } }, { spaces: 2 });
        }
    }
    async migrate() {
        const userPath = path.join(this.config.dataDir, 'plugins', 'package.json');
        if (!await deps_1.default.file.exists(userPath))
            return;
        let user = await deps_1.default.file.readJSON(userPath);
        if (!user.dependencies || user['cli-engine'])
            return;
        this.debug('migrating user plugins');
        user = await deps_1.default.file.readJSON(userPath);
        if (user['cli-engine'])
            return;
        for (let [name, tag] of Object.entries(user.dependencies)) {
            await this.addPlugin(name, tag);
        }
        user = await deps_1.default.file.readJSON(userPath);
        user['cli-engine'] = { schema: 1 };
        await deps_1.default.file.outputJSON(userPath, user);
    }
    async addPlugin(name, tag) {
        let plugins = (await this.manifest.get('plugins')) || {};
        plugins[name] = { tag };
        await this.manifest.set('plugins', plugins);
        await this.manifest.save();
    }
}
exports.UserPlugins = UserPlugins;
class UserPlugin extends plugin_1.Plugin {
    constructor(opts) {
        super(opts);
        this.type = 'user';
        this.tag = opts.tag || 'latest';
    }
}
exports.UserPlugin = UserPlugin;
